# Name:: FormatR
# Description:: Perl like formats for ruby
# Author:: Paul Rubel (rubel@crhc.uiuc.edu)
# Release:: 1.02
# Homepage:: http://www.crhc.uiuc.edu/~rubel/FormatR.html
# Date:: 21 December 2001
# License:: You can redistribute it and/or modify it under the same term as Ruby.
#           Copyright (c) 2001  Paul Rubel
# = To Test this code:
# You'll need perl installed and can try test_format.rb with no
# arguments. If nothing is amiss you should see Test successful and
# have a bunch of files of the type format_testfile laying around to
# examine.
#
# === Usage                                
# Class FormatR::Format in module FormatR provides perl like formats for ruby. 
# For a summary of the methods you're likely to need please see FormatR::Format.
# Formats are used to create output with a similar format but with changing
# values.                                  
#                                          
# For example:                             
#     require "format.rb"                  
#     include FormatR                      
#                                          
#     top_ex = <<DOT                       
#        Piggy Locations for @<< @#, @### 
#                          month, day, year
#                                          
#     Number: location              toe size
#     -------------------------------------------
#     DOT                                  
#                                          
#     ex = <<TOD                           
#     @)      @<<<<<<<<<<<<<<<<       @#.##
#     num,    location,             toe_size
#     TOD                                  
#                                          
#     top_fmt = Format.new (top_ex)        
#     body_fmt = Format.new (ex)           
#                                          
#     body_fmt.setTop(top_fmt)             
#     body_fmt.setPageLength(10)           
#     num = 1                              
#                                          
#     month = "Sep"                        
#     day = 18                             
#     year = 2001                          
#     ["Market", "Home", "Eating Roast Beef", "Having None", "On the way home"].each {|location|
#         toe_size = (num * 3.5)           
#         body_fmt.printFormat(binding)    
#         num += 1                         
#     }                                    
#                                          
#                                          
# When run produces the following output:  
#        Piggy Locations for Sep 18, 2001  
#                                          
#     Number: location              toe size
#     -------------------------------------------
#     1)      Market                   3.50
#     2)      Home                     7.00
#     3)      Eating Roast Beef       10.50
#     4)      Having None             14.00
#     5)      On the way home         17.50
#                                          
#                                          
# More examples are found in test_format.rb
#
# = Changes:
#
# ==1.02
# * Allow formats to be passed in as arrays of strings as well as just long strings
#
# * Added functionality so that if the first format on a page is too
#   long to fit on that page it will be printed partially with a
#   bottom. Perl seems to just print the whole thing and ignore the page
#   size in this case.
#
# * Fixed a bug where if your number didn't have a fractional part it
#   would crash if you used a format that need a fractional portion like @##.##
#
# * On the recommendation of Hugh Sasse added
#   finishPageWithoutFF(aBinding, io=@io) and 
#   finishPageWithFF(aBinding, io=@io) which will print out blank
#   lines until the end of the page and then print the bottom, with
#   and without a ^L. Only works on fixed sized bottoms.
#
# ==1.01
# * Moved to rdoc for generating documentation.
#
# ==1.00
# * Bottoms work iff you have a fixed size format and print out a
#   top afterwords. This means that you will only get a bottom if you
#   will print a top right after it so the last format page you print
#   won't have a bottom. It's impossible to figure out if you are
#   done with the format and therefore need to print the
#   bottom. Perhaps in a future release we can just take fixed sized
#   bottoms off the available size and get them to work that way.
# * Added support for Format.pageNumber()
# * Support ~ to be a space
# * Support ~ to suppress lines when the variables are empty
# * Support ~~ to repeat until the variables are empty
# * Support comments. If the first character in a line is a # the
#   line is a comment
# * Testing now compares against perl, it's a bit easier than
#   writing the tests manually.
# ==0.93
# * Added support for the ^ character to start a format
# 
# == 0.92
# * Added end of page characters and introduced line counts. 
#
# * Added the ability to manipulate the line count in case you write
#   to the file handle youself
#
# * Added format sizes. They just give the number of lines in the
#   current format. They don't try to iterate and get some total
#   count including tops and bottoms.
#
# = Incompatibilities/Issues
# * The first format printed on a page won't see any changes done on 
#   account of printing the top. This needs to be fixed.
#
# * If you use bottom be sure to check that you're happy with the
#   output. It doesn't currently work with variable sized bottoms. You
#   can use the finishPageWith{out}FF(...) methods to print out a
#   bottom if you're done printing but haven't finished a page.
#
# * Watch out for @#@??? as formats, see [ruby-talk:27782] and
#   [ruby-talk:27734]. This should be fixed in a future version of
#   ruby. The basic problem is that the here documents are equivalent
#   to "" and not '', they will evaluate variables in them. If this is
#   a problem be sure to just make a long string with '' and pass that
#   in. The next version of FormatR should suppory arrays as
#   constructor arguments.
#
# * Rounding seems to be broken in perl, if you try to print the following
#   format with 123.355 you won't get the same answer, you'll get 123.35 and
#   123.36. FormatR rounds "correctly" and plans to unless there is a
#   convincing reason not to.
#    format TEST_FORMAT =
#      ^#.### ^##.##
#    $num,  $num
#
#
# =To Do:
# *  Is there a better name than resetPage?
# *  From Hugh Sasse: The only other thing I wanted from Perl formats, which was not there,
#    was a means to set the maximum width, and create picture lines
#    computationally, so I could decide I wanted this and that on the left,
#    such and such on the right, and *the rest* (the middle) filled out with
#    some data without having to bang away on the < key for ages, hoping
#    I got the width right.
# *  Fix variable sized bottoms better. I'm not sure if this is
#    possible. You could try computing it first but this would cause
#    trouble if it depends upon the body format. I'm currently planning
#    to just live with fixed sized bottoms.

#
# =Thanks go to
# Hugh Sasse for some enlightening comments and suggestions on bottoms.

module FormatR

  # an exception that we can throw
  class FormatException < Exception
  end

  # This class holds a single block of text, either something
  # unchanging or a picture element of some format.
  class FormatEntry
    attr_accessor :val

    def initialize (val, space)
      @space = space
      @val = val
      if (!space)
        s = val.size - 1
        
        if (val =~ /[@^][<]{#{s},#{s}}/)
          @formatter = LeftFormatter.new (val)
        elsif (val =~ /[@^][>]{#{s},#{s}}/)     
          @formatter = RightFormatter.new (val)
        elsif (val =~ /[@^][\|]{#{s},#{s}}/)
          @formatter = CenterFormatter.new (val)
        elsif (val =~ /[@^](#*)([\.]{0,1})(#*)/)
          @formatter = NumberFormatter.new ($1, $2, $3)
        else
          raise FormatException.new(), "Malformed format entry \"#{@val}\""
        end
      end
    end
    
    # is this just unchanging characters
    def isUnchanging ()
      return @space
    end

    # give back the string passed through the appropriate formatter
    def formatString (string, var_name=nil, aBinding=nil)
      result = @formatter.formatString(string, var_name, aBinding)
      return result
    end
  end

  # This is the base class for all the formats, <,>,|, and # of the
  # @ or ^ persuasion. It keeps track of filled variables and the length
  # string should have.
  class Formatter
    def initialize (val)
      @len = val.size()
      @filled = false
      if (val =~ /\^.*/)
        @filled = true
      end
    end

    #if it's a filled field chop the displayed stuff off in the context given
    def changeVarValue (var_value, var_name, aBinding)
      result = var_value[0, @len]
      max_space = var_value[0,@len + 1].rindex (' ')
      if (max_space != nil)
        result = var_value[0,max_space]
      end
      s = "#{var_name} = #{var_name}[#{result.size()},#{var_value.size()}].gsub(/^\\s*/,'');" +
        "if #{var_name} == nil then #{var_name} = ''; end"
      eval (s, aBinding)
      return result
    end
    
    # return a formateed string of the correct length
    def formatString (var_value, var_name, binding)
      result = var_value[0,@len]
      if (! @filled)
        return result
      end
      return changeVarValue(var_value, var_name, binding)
    end


  end

  # this format doesn't care if it's a @ or an ^, it acts the same and doesn't chop things
  # used for @##.## formats
  class NumberFormatter < Formatter
    def initialize (wholeString, radix, fractionString)
      @whole = wholeString.size + 1 # for the '@'
      @fraction = fractionString.size
      @radix = radix.size #should always be 1
      @len = @whole + @fraction + @radix
    end
    
    # given a string that's a number spit it back with the right number of digits
    # and rounded the correct amount.
    def formatString (s, unused_var_name=nil, unused_aBinding=nil)
      num = s.split('.') # should this take into account internationalization?
      if (s.size == 1)
        return formatInt(s)
      end

      res = num[0]
      spaceLeft = @fraction + @radix
      if (res.size > @whole)
        spaceLeft = @len - res.size()
      end
      if (spaceLeft > 0)
        res += '.'  
        spaceLeft -= 1
      end
      res += getFract (num, spaceLeft) if (spaceLeft > 0)

      max = @len
      if (res.size > max)
        res = res[0,max]
      end
      res.rjust (max)
    end

    def formatInt (s)
      s.to_s.ljust (@len)
    end

    # what portion of the number is after the decimal point and should be printed
    def getFract (num, spaceLeft)
      num[1] = "" if (nil == num[1])
      @fraction.times {num[1] += '0'}
      fract = num[1][0,spaceLeft + 1]
      if (fract.size() >= spaceLeft + 1)
        if ((fract[spaceLeft,1].to_i) >= 5 )
          fract[spaceLeft - 1, 1] = ((fract[spaceLeft - 1, 1].to_i) + 1).to_s
        end
      end
      return fract[0,spaceLeft] 
    end
  end

  ## Format things that go to the left, ala <
  class LeftFormatter < Formatter
    def initialize (val)
      super
    end

    def formatString (s, var_name, binding)
      s = super 
      s.ljust(@len)
    end
  end

  ## Format things that go to the right, ala >
  class RightFormatter < Formatter
    def initialize (val)
      super
      @len = val.size()
    end

    def formatString (s, var_name, binding)
      s = super 
      s.rjust(@len)
    end
  end

  ## Format things that go to the center, ala |
  class CenterFormatter < Formatter
    def initialize (val)
      super
      @len = val.size()
    end

    def formatString (s, var_name, binding)
      s = super
      s.center(@len)
    end
  end


  # The class that exports the functionality that a user is interested in.
  class Format 
    public 
    # Set the IO that the format will be printed to, from stdout to a
    # file for example.
    attr_accessor  :io

    # Print out the specified format. You need to pass in a Binding
    # object for the variables that will be used in the format. This is
    # usually just a call to Kernel.binding. The next argument gives a
    # file handler to print to. This is useful so that a top or
    # bottom's output get written to the same place as the main format
    # is going even if their formats have a different io when they're
    # not attached.
    def printFormat (aBinding, io = @io)
      @binding = aBinding
      return printBodyFormat (io)
    end

    # When you don't want anymore on this page just fill it with blank
    # lines and print the bottom if it's there, print a ^L also. This
    # is good if you want to finish off the page but print more later
    # to the same file.
    def finishPageWithFF (aBinding, io = @io)
      finishPage (aBinding, false, io)
    end

    # When you don't want anymore on this page just fill it with blank
    # lines and print the bottom if it's there. Don't print a ^L at
    # the end. This is good if this will be the last page.
    def finishPageWithoutFF (aBinding, io = @io)
      finishPage (aBinding, true, io)
    end

    # Return how many times the top has been printed. You can use this
    # to number pages. An empty top can be used if you need the page
    # number but don't want to print a header
    def pageNumber()
      return @page_number
    end

    # how big is the format, useful to try to get bottom to work.
    # heaven help the person who tried bottom with a variable length
    # format
    def getSize ()
      @format_length
    end
    
    # If you want something to show up before the regular text of a
    # format you can specify it here. It will be printed once above the
    # format it is being set within. 
    def setTop (format)
      raise FormatException.new(), "recursive format not allowed" if (format == self)
      @top = format 
    end

    # Set a format to print at the end of a page. This is trickey and
    # you should be careful using it. It currently has problems on
    # short pages (at least). In order for a bottom to show up you need
    # to be ready to print a top. This means that formats less than a
    # page will not have bottoms.
    def setBottom (format)
      raise FormatException, "recursive format not allowed" if (format == self)
      @bottom = format
    end

    # Sets the number of lines on a page. If you don't want page breaks
    # set this to some large number that you hope you won't offset or
    # liberally use resetPage. The default is 60.
    def setPageLength(len)
      @page_length = len
      resetPage ()
    end

    # Sets the variable that says how many lines may be printed to the
    # maximum for the page which can be set using setPageLength (anInt).
    # Defaults to 60.
    def resetPage ()
      @lines_left = @page_length
    end

    # If you're writing to the file handle in another way than by
    # calling printFormat you can keep the pagination working using
    # this call to correctly keep track of lines.
    def addToLineCount (line_change)
      @lines_left += line_change
    end

    
    # Create a new format that will puts output to the given io
    # object. The output defaults to standard out. The format is a
    # string in the style of a perl format or an array of strings each
    # of which is a line of a perl format. The passed in format
    # contains multiple lines, picture lines and argument lines. A
    # picture line can contain any text but if it contains an at field
    # (@ followed by any number of <,>,| or a group of #'s of the
    # format #*.#* or #*) it must be followed by an argument line. The
    # arguments in the argument line are inserted in place of the at
    # fields in the picture line. Perl documentation for formats can
    # be found here:
    # http://www.cpan.org/doc/manual/html/pod/perlform.html
    # 
    # An example of a format is
    #  format = <<DOT
    #  Name: @<<<       @<<<<<
    #        first_name last_name
    #  DOT
    # 
    # This line specifies that when requested one line should be
    # printed and that it will say "Name: #{first_name} #{last_name}\n" but
    # that if either of those variables is longer than the length its
    # format the result will be truncated to the length of the format.
    # 
    # An at field specified as @<* specifies that the variable should
    # be left justified within the space allocated. @>* is right
    # justified, and @| is centered. #'s are used to print numbers and
    # can be used to set the number of digits after the decimal
    # point. However the whole number portion of an argument will
    # always be printed in its entirety even if it takes space set for
    # the fractional portion or even more space. If the fractional
    # portion is not long enough to fill the described space it will be
    # padded with 0s.
    def initialize (format, file = $stdout)
      @io = file
      @picture_lines = []
      @vars = []
      @top = @bottom = nil
      @page_length = 60
      @lines_left = @page_length
      @format_length = 0

      @buffered_lines = []
      @print_top = true
      @printed_a_body = false
      @print_bottom = false

      @page_number = 0
      lines = format
      if (format.type == String)
        lines = format.split( '\n' )
      end
      expected_vars = 0
      lines.each {|line|
        if (line =~ /^#.*/)
          #don't do anything, it's a comment
        elsif (0 != expected_vars)
          expected_vars = getVarLine(line, expected_vars)
        else
          expected_vars = getPictureLine(line, expected_vars)
          @format_length += 1
        end
      }
    end


    # Things you shouldn't have to deal with.
    private

    # When you don't want anymore on this page just fill it with blank
    # lines and print the bottom if it's there
    def finishPage (aBinding, suppressFF, io)
      @binding = aBinding
      tryOutputFormat (io)
      bottom_size = 0
      (bottom_size = @bottom.getSize()) if (@bottom)
      (@lines_left - bottom_size).times { io.puts ("")}
      @print_bottom = true
      tryPrintBottom (io, suppressFF)
    end
        
    # pull out the formatting
    def getPictureLine (line, expected_vars)
      num_vars = line.count('@') + line.count('^')
      if (num_vars != 0)
        expected_vars = num_vars
      else 
        @vars.push ([])
      end
      nonFormats = getNonFormats(line)
      formats = getFormats(line)
      a = FormatHolder.new()
      a.repeat = false
      a.repeat = true if (line =~ /^~~.*/)
      a.suppress = false
      a.suppress = true if (line =~ /^~.*/)
      nonFormats.each_index {|i|
        a.push( FormatEntry.new( nonFormats[i], true )) if ( nonFormats[i] )
        a.push( FormatEntry.new( formats[i], false )) if ( formats[i] )
      }
      @picture_lines.push(a)
      return expected_vars
    end

    # what variables shold be put into the picture line above
    def getVarLine (line, expected_vars)
      vars = line.split(',')
      if (vars.size != expected_vars)
        raise FormatException.new(),"malformed format, not enough variables provided. " +
          "Expected #{expected_vars} but received '#{line}'" 
      end
      @vars.push(vars)
      expected_vars = 0
      return expected_vars
    end

    # pull out each indifidual format from a line and return a list of
    # them
    def getFormats (line)
      last_found = line.size()
      output = []
      var_count = line.count('@') + line.count('^')
      var_count.times {|i|
        last_found = findFormatBefore(last_found, line, output)
      }
      output
    end

    # find a format before the position given in last_found and shove
    # it on the output
    def findFormatBefore (last_found, line, output)
      first_hat = line.rindex('^',last_found)
      first_at = line.rindex('@',last_found)
      first_hat = -1 if !first_hat 
      first_at  = -1 if !first_at
      first_index = (first_hat > first_at) ? first_hat : first_at
      first_char = (first_hat > first_at) ? '^' : '@'

      last_index = first_index
      line_section = line[(last_index + 1),(last_found - last_index)]
      # all the formats that we could have, blech this is ugly
      [/^(>+)[^>]*/, 
        /^(\|+)[^\|]*/,
        /^(#+\.{0,1}#*).*/,
        /^(#*\.{0,1}#+).*/,
        /^(<*)[^<]*/].each {|re|
        if (line_section =~ re)
          output.unshift(first_char + $1)
          last_found = (first_index - 1)
          return last_found
        end
      }
    end

    # pull out from a picture line the components of the line that aren't formats
    def getNonFormats (picture_line) 
      lines = picture_line.split(/[@^]/)
      output = []
      lines = lines.each {|element|
        #we don't really care if it's a @ or a ^, choose ^
        element = '@' + element
        element.gsub!(/^@#+\.{0,1}#*/, '')
        element.gsub!(/^@#*\.{0,1}#+/, '')
        element.gsub!(/^@>+/, '')
        element.gsub!(/^@\|+/, '')
        element.gsub!(/^@<*/, '')
        element.gsub!(/^@/, '')
        element.gsub!(/~/, ' ')
        output.push(element)
      }
      return output
    end

    ## print related functions

    # Try to save a line for outputting and perhaps a top and bottom.
    # we need the whole format to be able to print so buffer until we
    # get it
    def printLine (line, io, suppress = false)
      if (!suppress)
        line.gsub!(/\s*$/, "")
        @buffered_lines.push ("#{line}\n")
        tryPrintPartialPage(io) 
        tryPrintBottom (io)
        tryPrintTop (io) # this is a bug and should be printed first!
      end
    end

    # if the page is too short to hold even one format just print what
    # we can
    def tryPrintPartialPage (io)
      if (!@printed_a_body)
        bottom_size = 0
        bottom_size = @bottom.getSize() if (@bottom)
        if ((@buffered_lines.size() + bottom_size) == @lines_left)
          printBufferedLines (io)
          @print_bottom = true
        end
      end
    end

    # When we have a whole format try to print it, if there isn't
    # enough room we have to save it for later.
    def tryOutputFormat (io) 
      bottom_size = 0
      bottom_size = @bottom.getSize() if (@bottom)
      #puts "buffered size #{@buffered_lines.size} bottom #{bottom_size} left: #{@lines_left}"
      if ((@buffered_lines.size() + bottom_size) <= @lines_left)
        printBufferedLines (io)
      else
        @print_bottom = true
      end 
    end
    
    #print the buffered lines
    def printBufferedLines (io)
      io.puts (@buffered_lines)
      @lines_left -= (@buffered_lines.size())
      @buffered_lines = []
      @printed_a_body = true
    end

    # see if a top is the right thing to print
    def tryPrintTop (io)
      if (@print_top)
        @printed_a_body = false
      end
      if (@print_top && @top)
        @print_top = false
        @top.resetPage ()
        lines = @top.printFormat (@binding, io)
        @page_number += 1
        @lines_left = (@page_length - lines)
      end
    end

    #we have a bottom, even if it's only ^L, try to get this working!
    def tryPrintBottom (io, suppressLF = false)
      bottom_size = 0
      bottom_size = @bottom.getSize () if (@bottom)

      #this bottom is a mess if we have repeating lines
      if (@print_bottom)
        @print_bottom = false
        if (@bottom)
          @bottom.printFormat(@binding, io) 
          @bottom.resetPage () 
        end
        @lines_left = @page_length
        io.print '' unless suppressLF

        @print_top = true
        return true
      end
      return false
    end

    # The workhorse of printFormat, buffer the format contained and
    #  then try to print it out
    def printBodyFormat (io)
      @picture_lines.each_index {|i| # There isn't a 1-1 correspondance here anymore!
        line = @picture_lines[i]
        suppress = line.suppress()
        repeat = true
        while (repeat) 
          outputLine = ""
          vars = @vars[i].dup if @vars[i]
          line.each {|item|
            if (item.isUnchanging()) 
              outputLine << "#{item.val}"
            else
              begin
                to_eval = vars.shift
                raise FormatException.new(), 
                  "Not enough variables supplied to match format" if (nil == to_eval)
                s = eval (to_eval, @binding)
              rescue NameError
                raise NameError.new(), "cannot find variable '#{to_eval}'"
              end
              suppress = false if (("" != s) && suppress) 
              res = item.formatString(s.to_s, to_eval, @binding)
              outputLine << "#{res}"
            end
          }
          printLine (outputLine, io, suppress)
          if ((false == suppress) && line.repeat)
            suppress = line.suppress()
          else
            repeat = false
          end
        end #while
      }
      tryOutputFormat (io)
      return (@page_length - @lines_left)
    end

  end

  # a subclass of array that knows about ~ and ~~
  class FormatHolder < Array
    attr_accessor :suppress, :repeat
  end

end
